/// <reference types="node" />
import { DevContainerConfig, DevContainerFeature, VSCodeCustomizations } from './configuration';
import { Log } from '../spec-utils/log';
import { OCIFeatureRef, OCIManifest } from './containerFeaturesOCI';
export declare const V1_DEVCONTAINER_FEATURES_FILE_NAME = "devcontainer-features.json";
export declare const DEVCONTAINER_FEATURE_FILE_NAME = "devcontainer-feature.json";
export interface Feature {
    id: string;
    version?: string;
    name?: string;
    description?: string;
    cachePath?: string;
    internalVersion?: string;
    consecutiveId?: string;
    documentationURL?: string;
    licenseURL?: string;
    options?: Record<string, FeatureOption>;
    buildArg?: string;
    containerEnv?: Record<string, string>;
    mounts?: Mount[];
    init?: boolean;
    privileged?: boolean;
    capAdd?: string[];
    securityOpt?: string[];
    entrypoint?: string;
    include?: string[];
    exclude?: string[];
    value: boolean | string | Record<string, boolean | string | undefined>;
    included: boolean;
    customizations?: VSCodeCustomizations;
    installsAfter?: string[];
}
export declare type FeatureOption = {
    type: 'boolean';
    default?: boolean;
    description?: string;
} | {
    type: 'string';
    enum?: string[];
    default?: string;
    description?: string;
} | {
    type: 'string';
    proposals?: string[];
    default?: string;
    description?: string;
};
export interface Mount {
    type: 'bind' | 'volume';
    source: string;
    target: string;
    external?: boolean;
}
export declare function parseMount(str: string): Mount;
export declare type SourceInformation = LocalCacheSourceInformation | GithubSourceInformation | DirectTarballSourceInformation | FilePathSourceInformation | OCISourceInformation;
interface BaseSourceInformation {
    type: string;
    userFeatureId: string;
    userFeatureIdWithoutVersion?: string;
}
export interface LocalCacheSourceInformation extends BaseSourceInformation {
    type: 'local-cache';
}
export interface OCISourceInformation extends BaseSourceInformation {
    type: 'oci';
    featureRef: OCIFeatureRef;
    manifest: OCIManifest;
    userFeatureIdWithoutVersion: string;
}
export interface DirectTarballSourceInformation extends BaseSourceInformation {
    type: 'direct-tarball';
    tarballUri: string;
}
export interface FilePathSourceInformation extends BaseSourceInformation {
    type: 'file-path';
    resolvedFilePath: string;
}
export interface GithubSourceInformation extends BaseSourceInformation {
    type: 'github-repo';
    apiUri: string;
    unauthenticatedUri: string;
    owner: string;
    repo: string;
    isLatest: boolean;
    tag?: string;
    ref?: string;
    sha?: string;
    userFeatureIdWithoutVersion: string;
}
export interface GithubSourceInformationInput {
    owner: string;
    repo: string;
    ref?: string;
    sha?: string;
    tag?: string;
}
export interface FeatureSet {
    features: Feature[];
    internalVersion?: string;
    sourceInformation: SourceInformation;
}
export interface FeaturesConfig {
    featureSets: FeatureSet[];
    dstFolder?: string;
}
export interface GitHubApiReleaseInfo {
    assets: GithubApiReleaseAsset[];
    name: string;
    tag_name: string;
}
export interface GithubApiReleaseAsset {
    url: string;
    name: string;
    content_type: string;
    size: number;
    download_count: number;
    updated_at: string;
}
export interface CollapsedFeaturesConfig {
    allFeatures: Feature[];
}
export declare function collapseFeaturesConfig(original: FeaturesConfig): CollapsedFeaturesConfig;
export declare const multiStageBuildExploration = false;
export declare function getContainerFeaturesFolder(_extensionPath: string | {
    distFolder: string;
}): string;
export declare function getSourceInfoString(srcInfo: SourceInformation): string;
export declare function getContainerFeaturesBaseDockerFile(): string;
export declare function getFeatureInstallWrapperScript(feature: Feature, featureSet: FeatureSet, options: string[]): string;
export declare function getFeatureLayers(featuresConfig: FeaturesConfig): string;
export declare function generateContainerEnvs(feature: Feature): string;
export declare function loadFeaturesJson(jsonBuffer: Buffer, filePath: string, output: Log): Promise<FeatureSet | undefined>;
export declare function loadV1FeaturesJsonFromDisk(pathToDirectory: string, output: Log): Promise<FeatureSet | undefined>;
export declare function generateFeaturesConfig(params: {
    extensionPath: string;
    cwd: string;
    output: Log;
    env: NodeJS.ProcessEnv;
    skipFeatureAutoMapping: boolean;
}, dstFolder: string, config: DevContainerConfig, getLocalFeaturesFolder: (d: string) => string): Promise<FeaturesConfig | undefined>;
export declare function getFeatureIdType(output: Log, env: NodeJS.ProcessEnv, userFeatureId: string): Promise<{
    type: string;
    manifest: undefined;
} | {
    type: string;
    manifest: OCIManifest;
}>;
export declare function getBackwardCompatibleFeatureId(output: Log, id: string): string;
export declare function processFeatureIdentifier(output: Log, configPath: string, _workspaceRoot: string, env: NodeJS.ProcessEnv, userFeature: DevContainerFeature, skipFeatureAutoMapping?: boolean): Promise<FeatureSet | undefined>;
export declare function getFeatureMainProperty(feature: Feature): "version" | undefined;
export declare function getFeatureMainValue(feature: Feature): string | boolean | undefined;
export declare function getFeatureValueObject(feature: Feature): Record<string, string | boolean | undefined>;
export {};
