"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.packageCollection = exports.packageSingleFeatureOrTemplate = exports.prepPackageCommand = exports.OCICollectionFileName = void 0;
const tar_1 = __importDefault(require("tar"));
const pfs_1 = require("../../spec-utils/pfs");
const log_1 = require("../../spec-utils/log");
const path_1 = __importDefault(require("path"));
exports.OCICollectionFileName = 'devcontainer-collection.json';
async function prepPackageCommand(args, collectionType) {
    const { cliHost, targetFolder, outputDir, forceCleanOutputDir, output, disposables } = args;
    const targetFolderResolved = cliHost.path.resolve(targetFolder);
    if (!(await (0, pfs_1.isLocalFolder)(targetFolderResolved))) {
        throw new Error(`Target folder '${targetFolderResolved}' does not exist`);
    }
    const outputDirResolved = cliHost.path.resolve(outputDir);
    if (await (0, pfs_1.isLocalFolder)(outputDirResolved)) {
        // Output dir exists. Delete it automatically if '-f' is true
        if (forceCleanOutputDir) {
            await (0, pfs_1.rmLocal)(outputDirResolved, { recursive: true, force: true });
        }
        else {
            output.write(`(!) ERR: Output directory '${outputDirResolved}' already exists. Manually delete, or pass '-f' to continue.`, log_1.LogLevel.Error);
            process.exit(1);
        }
    }
    // Detect if we're packaging a collection or a single feature/template
    const isValidFolder = await (0, pfs_1.isLocalFolder)(cliHost.path.join(targetFolderResolved));
    const isSingle = await (0, pfs_1.isLocalFile)(cliHost.path.join(targetFolderResolved, `devcontainer-${collectionType}.json`));
    if (!isValidFolder) {
        throw new Error(`Target folder '${targetFolderResolved}' does not exist`);
    }
    // Generate output folder.
    await (0, pfs_1.mkdirpLocal)(outputDirResolved);
    return {
        cliHost,
        targetFolder: targetFolderResolved,
        outputDir: outputDirResolved,
        forceCleanOutputDir,
        output,
        disposables,
        isSingle
    };
}
exports.prepPackageCommand = prepPackageCommand;
async function tarDirectory(folder, archiveName, outputDir) {
    return new Promise((resolve) => resolve(tar_1.default.create({ file: path_1.default.join(outputDir, archiveName), cwd: folder }, ['.'])));
}
const getArchiveName = (f, collectionType) => `devcontainer-${collectionType}-${f}.tgz`;
async function packageSingleFeatureOrTemplate(args, collectionType) {
    const { output, targetFolder, outputDir } = args;
    let metadatas = [];
    const jsonPath = path_1.default.join(targetFolder, `devcontainer-${collectionType}.json`);
    const metadata = JSON.parse(await (0, pfs_1.readLocalFile)(jsonPath, 'utf-8'));
    if (!metadata.id || !metadata.version) {
        output.write(`${collectionType} is missing an id or version in its devcontainer-${collectionType}.json`, log_1.LogLevel.Error);
        return;
    }
    const archiveName = getArchiveName(metadata.id, collectionType);
    await tarDirectory(targetFolder, archiveName, outputDir);
    output.write(`Packaged ${collectionType} '${metadata.id}'`, log_1.LogLevel.Info);
    metadatas.push(metadata);
    return metadatas;
}
exports.packageSingleFeatureOrTemplate = packageSingleFeatureOrTemplate;
// Packages collection of Features or Templates
async function packageCollection(args, collectionType) {
    const { output, targetFolder: srcFolder, outputDir } = args;
    const collectionDirs = await (0, pfs_1.readLocalDir)(srcFolder);
    let metadatas = [];
    for await (const c of collectionDirs) {
        output.write(`Processing ${collectionType}: ${c}...`, log_1.LogLevel.Info);
        if (!c.startsWith('.')) {
            const folder = path_1.default.join(srcFolder, c);
            const archiveName = getArchiveName(c, collectionType);
            // Validate minimal folder structure
            const devcontainerJsonName = `devcontainer-${collectionType}.json`;
            const jsonPath = path_1.default.join(folder, devcontainerJsonName);
            if (!(0, pfs_1.isLocalFile)(jsonPath)) {
                output.write(`${collectionType} '${c}' is missing a ${devcontainerJsonName}`, log_1.LogLevel.Error);
                return;
            }
            if (collectionType === 'feature') {
                const installShPath = path_1.default.join(folder, 'install.sh');
                if (!(0, pfs_1.isLocalFile)(installShPath)) {
                    output.write(`Feature '${c}' is missing an install.sh`, log_1.LogLevel.Error);
                    return;
                }
            }
            else if (collectionType === 'template') {
                const devcontainerFile = path_1.default.join(folder, '.devcontainer.json');
                const devcontainerFileWithinDevcontainerFolder = path_1.default.join(folder, '.devcontainer/devcontainer.json');
                if (!(0, pfs_1.isLocalFile)(devcontainerFile) && !(0, pfs_1.isLocalFile)(devcontainerFileWithinDevcontainerFolder)) {
                    output.write(`Template '${c}' is missing a devcontainer.json`, log_1.LogLevel.Error);
                    return;
                }
            }
            await tarDirectory(folder, archiveName, outputDir);
            const metadata = JSON.parse(await (0, pfs_1.readLocalFile)(jsonPath, 'utf-8'));
            if (!metadata.id || !metadata.version) {
                output.write(`${collectionType} '${c}' is missing an id or version in its ${devcontainerJsonName}`, log_1.LogLevel.Error);
                return;
            }
            metadatas.push(metadata);
        }
    }
    if (metadatas.length === 0) {
        return;
    }
    output.write(`Packaged ${metadatas.length} ${collectionType}!`, log_1.LogLevel.Info);
    return metadatas;
}
exports.packageCollection = packageCollection;
//# sourceMappingURL=packageCommandImpl.js.map