"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.ensureDockerfileHasFinalStageName = exports.findBaseImage = exports.findUserStatement = exports.extractDockerfile = exports.RemoteDocuments = exports.fileDocuments = exports.createDocuments = exports.CLIHostDocuments = exports.parentURI = exports.uriToFsPath = exports.resolveConfigFilePath = exports.isDockerFileConfig = exports.getDockerfilePath = exports.getConfigFilePath = void 0;
var configuration_1 = require("../spec-configuration/configuration");
Object.defineProperty(exports, "getConfigFilePath", { enumerable: true, get: function () { return configuration_1.getConfigFilePath; } });
Object.defineProperty(exports, "getDockerfilePath", { enumerable: true, get: function () { return configuration_1.getDockerfilePath; } });
Object.defineProperty(exports, "isDockerFileConfig", { enumerable: true, get: function () { return configuration_1.isDockerFileConfig; } });
Object.defineProperty(exports, "resolveConfigFilePath", { enumerable: true, get: function () { return configuration_1.resolveConfigFilePath; } });
var configurationCommonUtils_1 = require("../spec-configuration/configurationCommonUtils");
Object.defineProperty(exports, "uriToFsPath", { enumerable: true, get: function () { return configurationCommonUtils_1.uriToFsPath; } });
Object.defineProperty(exports, "parentURI", { enumerable: true, get: function () { return configurationCommonUtils_1.parentURI; } });
var editableFiles_1 = require("../spec-configuration/editableFiles");
Object.defineProperty(exports, "CLIHostDocuments", { enumerable: true, get: function () { return editableFiles_1.CLIHostDocuments; } });
Object.defineProperty(exports, "createDocuments", { enumerable: true, get: function () { return editableFiles_1.createDocuments; } });
Object.defineProperty(exports, "fileDocuments", { enumerable: true, get: function () { return editableFiles_1.fileDocuments; } });
Object.defineProperty(exports, "RemoteDocuments", { enumerable: true, get: function () { return editableFiles_1.RemoteDocuments; } });
const findFromLines = new RegExp(/^(?<line>\s*FROM.*)/, 'gm');
const parseFromLine = /FROM\s+(?<platform>--platform=\S+\s+)?(?<image>\S+)(\s+[Aa][Ss]\s+(?<label>[^\s]+))?/;
const fromStatement = /^\s*FROM\s+(?<platform>--platform=\S+\s+)?(?<image>\S+)(\s+[Aa][Ss]\s+(?<label>[^\s]+))?/m;
const userStatements = /^\s*USER\s+(?<user>\S+)/gm;
const argStatementsWithValue = /^\s*ARG\s+(?<name>\S+)=("(?<value1>\S+)"|(?<value2>\S+))/gm;
function extractDockerfile(dockerfile) {
    const fromStatementsAhead = /(?=^\s*FROM)/gm;
    const parts = dockerfile.split(fromStatementsAhead);
    const preambleStr = fromStatementsAhead.test(parts[0] || '') ? '' : parts.shift();
    const stageStrs = parts;
    const stages = stageStrs.map(stageStr => {
        var _a;
        return ({
            from: ((_a = fromStatement.exec(stageStr)) === null || _a === void 0 ? void 0 : _a.groups) || { image: 'unknown' },
            args: extractArgs(stageStr),
            users: [...stageStr.matchAll(userStatements)].map(m => m.groups.user),
        });
    });
    return {
        preamble: {
            args: extractArgs(preambleStr),
        },
        stages,
        stagesByLabel: stages.reduce((obj, stage) => {
            if (stage.from.label) {
                obj[stage.from.label] = stage;
            }
            return obj;
        }, {}),
    };
}
exports.extractDockerfile = extractDockerfile;
function findUserStatement(dockerfile, buildArgs, target) {
    let stage = target ? dockerfile.stagesByLabel[target] : dockerfile.stages[dockerfile.stages.length - 1];
    const seen = new Set();
    while (stage) {
        if (seen.has(stage)) {
            return undefined;
        }
        seen.add(stage);
        if (stage.users.length) {
            return replaceArgs(stage.users[stage.users.length - 1], { ...stage.args, ...buildArgs });
        }
        const image = replaceArgs(stage.from.image, { ...dockerfile.preamble.args, ...buildArgs });
        stage = dockerfile.stagesByLabel[image];
    }
    return undefined;
}
exports.findUserStatement = findUserStatement;
function findBaseImage(dockerfile, buildArgs, target) {
    let stage = target ? dockerfile.stagesByLabel[target] : dockerfile.stages[dockerfile.stages.length - 1];
    const seen = new Set();
    while (stage) {
        if (seen.has(stage)) {
            return undefined;
        }
        seen.add(stage);
        const image = replaceArgs(stage.from.image, { ...dockerfile.preamble.args, ...buildArgs });
        const nextStage = dockerfile.stagesByLabel[image];
        if (!nextStage) {
            return image;
        }
        stage = nextStage;
    }
    return undefined;
}
exports.findBaseImage = findBaseImage;
function extractArgs(stageStr) {
    return [...stageStr.matchAll(argStatementsWithValue)]
        .reduce((obj, match) => {
        const groups = match.groups;
        obj[groups.name] = groups.value1 || groups.value2;
        return obj;
    }, {});
}
function replaceArgs(str, args) {
    return Object.keys(args)
        .sort((a, b) => b.length - a.length) // Sort by length to replace longest first.
        .reduce((current, arg) => current.replace(new RegExp(`\\$${arg}|\\$\\{${arg}\\}`, 'g'), args[arg]), str);
}
// not expected to be called externally (exposed for testing)
function ensureDockerfileHasFinalStageName(dockerfile, defaultLastStageName) {
    var _a, _b;
    // Find the last line that starts with "FROM" (possibly preceeded by white-space)
    const fromLines = [...dockerfile.matchAll(findFromLines)];
    const lastFromLineMatch = fromLines[fromLines.length - 1];
    const lastFromLine = (_a = lastFromLineMatch.groups) === null || _a === void 0 ? void 0 : _a.line;
    // Test for "FROM [--platform=someplat] base [as label]"
    // That is, match against optional platform and label
    const fromMatch = lastFromLine.match(parseFromLine);
    if (!fromMatch) {
        throw new Error('Error parsing Dockerfile: failed to parse final FROM line');
    }
    if ((_b = fromMatch.groups) === null || _b === void 0 ? void 0 : _b.label) {
        return {
            lastStageName: fromMatch.groups.label,
            modifiedDockerfile: undefined,
        };
    }
    // Last stage doesn't have a name, so modify the Dockerfile to set the name to defaultLastStageName
    const lastLineStartIndex = lastFromLineMatch.index + fromMatch.index;
    const lastLineEndIndex = lastLineStartIndex + lastFromLine.length;
    const matchedFromText = fromMatch[0];
    let modifiedDockerfile = dockerfile.slice(0, lastLineStartIndex + matchedFromText.length);
    modifiedDockerfile += ` AS ${defaultLastStageName}`;
    const remainingFromLineLength = lastFromLine.length - matchedFromText.length;
    modifiedDockerfile += dockerfile.slice(lastLineEndIndex - remainingFromLineLength);
    return { lastStageName: defaultLastStageName, modifiedDockerfile: modifiedDockerfile };
}
exports.ensureDockerfileHasFinalStageName = ensureDockerfileHasFinalStageName;
//# sourceMappingURL=dockerfileUtils.js.map