"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.doFeaturesTestCommand = void 0;
const path_1 = __importDefault(require("path"));
const chalk_1 = __importDefault(require("chalk"));
const os_1 = require("os");
const devContainers_1 = require("../devContainers");
const devContainersSpecCLI_1 = require("../devContainersSpecCLI");
const utils_1 = require("./utils");
const pfs_1 = require("../../spec-utils/pfs");
const TEST_LIBRARY_SCRIPT_NAME = 'dev-container-features-test-lib';
function fail(msg) {
    log(msg, { prefix: '[-]', stderr: true });
    process.exit(1);
}
function log(msg, options) {
    const prefix = (options === null || options === void 0 ? void 0 : options.prefix) || '> ';
    const output = `${(options === null || options === void 0 ? void 0 : options.omitPrefix) ? '' : `${prefix} `}${msg}\n`;
    if (options === null || options === void 0 ? void 0 : options.stderr) {
        process.stderr.write(chalk_1.default.red(output));
    }
    else if (options === null || options === void 0 ? void 0 : options.info) {
        process.stdout.write(chalk_1.default.bold.blue(output));
    }
    else {
        process.stdout.write(chalk_1.default.blue(output));
    }
}
async function doFeaturesTestCommand(args) {
    const { pkg, globalScenariosOnly, features, collectionFolder, cliHost } = args;
    process.stdout.write(`
┌ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ┐
|    dev container 'features' |   
│           v${pkg.version}           │
└ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ─ ┘\n\n`);
    const srcDir = `${collectionFolder}/src`;
    const testsDir = `${collectionFolder}/test`;
    if (!await cliHost.isFolder(srcDir) || !await cliHost.isFolder(testsDir)) {
        fail(`Folder '${collectionFolder}' does not contain the required 'src' and 'test' folders.`);
    }
    let testResults = [];
    if (globalScenariosOnly) {
        await runGlobalFeatureTests(args, testResults);
    }
    else {
        await runFeatureTests(args, testResults);
        // If any features were explictily set to run,
        // we know we don't want to run the global tests.
        if (!features) {
            await runGlobalFeatureTests(args, testResults);
        }
    }
    // Pretty-print test results and exit with 0 or 1 exit code.
    return analyzeTestResults(testResults);
}
exports.doFeaturesTestCommand = doFeaturesTestCommand;
async function runGlobalFeatureTests(args, testResults = []) {
    const { collectionFolder } = args;
    const globalTestsFolder = `${collectionFolder}/test/_global`;
    log(`Scenarios:         ${globalTestsFolder}\n`, { prefix: '\n📊', info: true });
    testResults = await doScenario(globalTestsFolder, args, testResults);
    if (!testResults) {
        fail(`Failed to run scenarios in ${globalTestsFolder}`);
        return []; // We never reach here, we exit via fail().
    }
    return testResults;
}
async function runFeatureTests(args, testResults = []) {
    const { baseImage, collectionFolder, remoteUser, cliHost, skipAutogenerated, skipScenarios } = args;
    let { features } = args;
    const testsDir = `${collectionFolder}/test`;
    log(`baseImage:         ${baseImage}`);
    log(`Target Folder:     ${collectionFolder}`);
    // Parse comma separated list of features
    // If a set of '--features' isn't specified, run all features with a 'test' subfolder in random order.
    if (!features) {
        // Auto-detect
        features =
            (await cliHost.readDir(testsDir))
                .filter(f => f !== '_global'); // Exclude any folder named '_global'
        if (features.length === 0) {
            fail(`No features specified and no test folders found in '${testsDir}'`);
        }
    }
    log(`features:          ${features.join(', ')}`);
    let workspaceFolder = undefined;
    let params = undefined;
    if (!skipAutogenerated) {
        // 1. Generate temporary project with 'baseImage' and all the 'features..'
        workspaceFolder = await generateDefaultProjectFromFeatures(cliHost, baseImage, collectionFolder, features, remoteUser);
        params = await generateDockerParams(workspaceFolder, args);
        await createContainerFromWorkingDirectory(params, workspaceFolder, args);
    }
    log('Starting test(s)...\n', { prefix: '\n🏃', info: true });
    // 3. Exec default 'test.sh' script for each feature, in the provided order.
    //    Also exec a test's test scenarios, if a scenarios.json is present in the feature's test folder.
    for (const feature of features) {
        log(`Starting '${feature}' tests...`, { prefix: '🧪' });
        const featureTestFolder = path_1.default.join(collectionFolder, 'test', feature);
        if (!skipAutogenerated) {
            if (!workspaceFolder || !params) {
                fail('Uninitialized workspaceFolder or params');
                return [];
            }
            await doRunAutoTest(feature, workspaceFolder, params, featureTestFolder, args, testResults);
        }
        // If there is a feature-scoped 'scenarios.json' with additional tests, also exec those.
        // Pass  'testResults' array reference in to capture results.
        if (!skipScenarios) {
            log(`Executing scenarios for feature '${feature}'...`, { prefix: '🧪' });
            await doScenario(featureTestFolder, args, testResults);
        }
        if (!testResults) {
            fail(`Failed to run tests`);
            return []; // We never reach here, we exit via fail().
        }
    }
    return testResults;
}
async function doRunAutoTest(feature, workspaceFolder, params, featureTestFolder, args, testResults = []) {
    const { cliHost } = args;
    const testScriptPath = path_1.default.join(featureTestFolder, 'test.sh');
    if (!(await cliHost.isFile(testScriptPath))) {
        fail(`Could not find test.sh script at ${testScriptPath}`);
    }
    // Move the test script into the workspaceFolder
    const testScript = await cliHost.readFile(testScriptPath);
    const remoteTestScriptName = `${feature}-test-${Date.now()}.sh`;
    await cliHost.writeFile(`${workspaceFolder}/${remoteTestScriptName}`, testScript);
    // Move the test library script into the workspaceFolder
    await cliHost.writeFile(`${workspaceFolder}/${TEST_LIBRARY_SCRIPT_NAME}`, Buffer.from(utils_1.testLibraryScript));
    // Execute Test
    const result = await execTest(params, remoteTestScriptName, workspaceFolder);
    testResults.push({
        testName: feature,
        result,
    });
    return testResults;
}
async function doScenario(pathToTestDir, args, testResults = []) {
    const { collectionFolder, cliHost } = args;
    const scenariosPath = path_1.default.join(pathToTestDir, 'scenarios.json');
    if (!(await cliHost.isFile(scenariosPath))) {
        log(`No scenario file found at '${scenariosPath}'. Skipping...`, { prefix: '⚠️', });
        return testResults;
    }
    // Read in scenarios.json
    const scenariosBuffer = await cliHost.readFile(scenariosPath);
    // Parse to json
    let scenarios = {};
    try {
        scenarios = JSON.parse(scenariosBuffer.toString());
    }
    catch (e) {
        fail(`Failed to parse scenarios.json:  ${e.message}`);
        return []; // We never reach here, we exit via fail().
    }
    // For EACH scenario: Spin up a container and exec the scenario test script
    for (const [scenarioName, scenarioConfig] of Object.entries(scenarios)) {
        log(`Running scenario:  ${scenarioName}`);
        // Check if we have a scenario test script, otherwise skip.
        const scenarioTestScript = path_1.default.join(pathToTestDir, `${scenarioName}.sh`);
        if (!(await cliHost.isFile(scenarioTestScript))) {
            log(`No scenario test script found at path ${scenarioTestScript}, skipping scenario...`);
            continue;
        }
        // Create Container
        const workspaceFolder = await generateProjectFromScenario(cliHost, collectionFolder, scenarioName, scenarioConfig);
        const params = await generateDockerParams(workspaceFolder, args);
        await createContainerFromWorkingDirectory(params, workspaceFolder, args);
        // Execute test script
        // Move the test script into the workspaceFolder
        const testScript = await cliHost.readFile(scenarioTestScript);
        const remoteTestScriptName = `${scenarioName}-test-${Date.now()}.sh`;
        await cliHost.writeFile(`${workspaceFolder}/${remoteTestScriptName}`, testScript);
        // Move the test library script into the workspaceFolder
        await cliHost.writeFile(`${workspaceFolder}/${TEST_LIBRARY_SCRIPT_NAME}`, Buffer.from(utils_1.testLibraryScript));
        // Execute Test
        testResults.push({
            testName: scenarioName,
            result: await execTest(params, remoteTestScriptName, workspaceFolder)
        });
    }
    return testResults;
}
function analyzeTestResults(testResults) {
    if (!testResults) {
        fail('No test results found!');
    }
    // 4. Print results
    // NOTE: 0 tests means allPassed == true.
    const allPassed = testResults.every((x) => x.result);
    process.stdout.write('\n\n\n');
    log('================== TEST REPORT ==================', { 'info': true, 'prefix': ' ' });
    testResults.forEach(t => {
        if (t.result) {
            log(`Passed:      '${t.testName}'`, { 'prefix': '✅', 'info': true });
        }
        else {
            log(`Failed:      '${t.testName}'`, { 'prefix': '❌', 'info': true });
        }
    });
    process.stdout.write('\n');
    return allPassed ? 0 : 1;
}
const devcontainerTemplate = `
{
	"image": "#{IMAGE}",
	"features": {
		#{FEATURES}
	},
	"remoteUser": "#{REMOTE_USER}"
}`;
async function createContainerFromWorkingDirectory(params, workspaceFolder, args) {
    const { quiet, remoteUser, disposables } = args;
    log(`workspaceFolder:   ${workspaceFolder}`);
    // 2. Use  'devcontainer-cli up'  to build and start a container
    log('Building test container...\n', { prefix: '\n⏳', info: true });
    const launchResult = await launchProject(params, workspaceFolder, quiet, disposables);
    if (!launchResult || !launchResult.containerId) {
        fail('Failed to launch container');
        return;
    }
    const { containerId } = launchResult;
    log(`Launched container.`, { prefix: '\n🚀', info: true });
    log(`containerId:          ${containerId}`);
    log(`remoteUser:           ${remoteUser}`);
    return launchResult;
}
async function createTempDevcontainerFolder(cliHost) {
    const systemTmpDir = (0, os_1.tmpdir)();
    const tmpFolder = path_1.default.join(systemTmpDir, 'devcontainercli', 'container-features-test', Date.now().toString());
    await cliHost.mkdirp(`${tmpFolder}/.devcontainer`);
    return tmpFolder;
}
async function generateDefaultProjectFromFeatures(cliHost, baseImage, collectionsDirectory, featuresToTest, remoteUser) {
    const tmpFolder = await createTempDevcontainerFolder(cliHost);
    const features = featuresToTest
        .map((x) => `"./${x}": {}`)
        .join(',\n');
    for (const featureId of featuresToTest) {
        // Copy the feature source code to the temp folder
        const pathToFeatureSource = `${collectionsDirectory}/src/${featureId}`;
        await (0, pfs_1.cpDirectoryLocal)(pathToFeatureSource, `${tmpFolder}/.devcontainer/${featureId}`);
    }
    let template = devcontainerTemplate
        .replace('#{IMAGE}', baseImage)
        .replace('#{FEATURES}', features)
        .replace('#{REMOTE_USER}', remoteUser);
    await cliHost.writeFile(`${tmpFolder}/.devcontainer/devcontainer.json`, Buffer.from(template));
    return tmpFolder;
}
async function generateProjectFromScenario(cliHost, collectionsDirectory, scenarioId, scenarioObject) {
    const tmpFolder = await createTempDevcontainerFolder(cliHost);
    let features = scenarioObject.features;
    if (!scenarioObject || !features) {
        fail(`Scenario '${scenarioId}' is missing features!`);
        return ''; // Exits in the 'fail()' before this line is reached.
    }
    // Prefix the local path to the collections directory
    let updatedFeatures = {};
    for (const [featureId, featureValue] of Object.entries(features)) {
        // Copy the feature source code to the temp folder
        const pathToFeatureSource = `${collectionsDirectory}/src/${featureId}`;
        await (0, pfs_1.cpDirectoryLocal)(pathToFeatureSource, `${tmpFolder}/.devcontainer/${featureId}`);
        // Reference Feature in the devcontainer.json
        updatedFeatures[`./${featureId}`] = featureValue;
    }
    scenarioObject.features = updatedFeatures;
    await cliHost.writeFile(`${tmpFolder}/.devcontainer/devcontainer.json`, Buffer.from(JSON.stringify(scenarioObject)));
    // tmpFolder will serve as our auto-generated 'workingFolder'
    return tmpFolder;
}
async function launchProject(params, workspaceFolder, quiet, disposables) {
    var _a;
    const { common } = params;
    let response = {};
    const options = {
        ...utils_1.staticProvisionParams,
        workspaceFolder,
        logLevel: common.getLogLevel(),
        mountWorkspaceGitRoot: true,
        idLabels: [
            `devcontainer.local_folder=${workspaceFolder}`
        ],
        remoteEnv: common.remoteEnv,
        skipFeatureAutoMapping: common.skipFeatureAutoMapping,
        experimentalImageMetadata: common.experimentalImageMetadata,
        log: text => quiet ? null : process.stderr.write(text),
    };
    try {
        if (quiet) {
            // Launch container but don't await it to reduce output noise
            let isResolved = false;
            const p = (0, devContainers_1.launch)(options, disposables);
            p.then(function (res) {
                process.stdout.write('\n');
                response = res;
                isResolved = true;
            });
            while (!isResolved) {
                // Just so visual progress with dots
                process.stdout.write('.');
                await new Promise((resolve) => setTimeout(resolve, 500));
            }
        }
        else {
            // Stream all the container setup logs.
            response = await (0, devContainers_1.launch)(options, disposables);
        }
        return {
            ...response,
            disposables,
        };
    }
    catch (e) {
        fail(`Failed to launch container:\n\n${(_a = e === null || e === void 0 ? void 0 : e.message) !== null && _a !== void 0 ? _a : 'Unknown error'}`);
        return response; // `fail` exits before we return this.
    }
}
async function execTest(params, remoteTestScriptName, workspaceFolder) {
    let cmd = 'chmod';
    let args = ['777', `./${remoteTestScriptName}`, `./${TEST_LIBRARY_SCRIPT_NAME}`];
    await exec(params, cmd, args, workspaceFolder);
    cmd = `./${remoteTestScriptName}`;
    args = [];
    return await exec(params, cmd, args, workspaceFolder);
}
async function exec(_params, cmd, args, workspaceFolder) {
    const execArgs = {
        ...utils_1.staticExecParams,
        'workspace-folder': workspaceFolder,
        'skip-feature-auto-mapping': false,
        'experimental-image-metadata': false,
        cmd,
        args,
        _: [
            cmd,
            ...args
        ]
    };
    const result = await (0, devContainersSpecCLI_1.doExec)(execArgs);
    return (result.outcome === 'success');
}
async function generateDockerParams(workspaceFolder, args) {
    const { logLevel, quiet, disposables } = args;
    return await (0, devContainers_1.createDockerParams)({
        workspaceFolder,
        dockerPath: undefined,
        dockerComposePath: undefined,
        containerDataFolder: undefined,
        containerSystemDataFolder: undefined,
        mountWorkspaceGitRoot: false,
        idLabels: [],
        configFile: undefined,
        overrideConfigFile: undefined,
        logLevel,
        logFormat: 'text',
        log: text => quiet ? null : process.stderr.write(text),
        terminalDimensions: undefined,
        defaultUserEnvProbe: 'loginInteractiveShell',
        removeExistingContainer: false,
        buildNoCache: false,
        expectExistingContainer: false,
        postCreateEnabled: false,
        skipNonBlocking: false,
        prebuild: false,
        persistedFolder: undefined,
        additionalMounts: [],
        updateRemoteUserUIDDefault: 'never',
        remoteEnv: {},
        additionalCacheFroms: [],
        omitLoggerHeader: true,
        useBuildKit: 'auto',
        buildxPlatform: undefined,
        buildxPush: false,
        skipFeatureAutoMapping: false,
        skipPostAttach: false,
        experimentalImageMetadata: false,
    }, disposables);
}
//# sourceMappingURL=testCommandImpl.js.map