"use strict";
/*---------------------------------------------------------------------------------------------
 *  Copyright (c) Microsoft Corporation. All rights reserved.
 *  Licensed under the MIT License. See License.txt in the project root for license information.
 *--------------------------------------------------------------------------------------------*/
Object.defineProperty(exports, "__esModule", { value: true });
exports.getDevcontainerMetadataLabel = exports.getImageMetadata = exports.getImageMetadataFromContainer = exports.imageMetadataLabel = exports.internalGetImageBuildInfoFromDockerfile = exports.getImageBuildInfoFromDockerfile = exports.getImageBuildInfoFromImage = exports.getImageBuildInfo = exports.getDevcontainerMetadata = exports.mergeConfiguration = void 0;
const errors_1 = require("../spec-common/errors");
const configuration_1 = require("../spec-configuration/configuration");
const dockerCompose_1 = require("./dockerCompose");
const dockerfileUtils_1 = require("./dockerfileUtils");
const utils_1 = require("./utils");
const pickConfigProperties = [
    'onCreateCommand',
    'updateContentCommand',
    'postCreateCommand',
    'postStartCommand',
    'postAttachCommand',
    'waitFor',
    'customizations',
    'mounts',
    'containerEnv',
    'containerUser',
    'init',
    'privileged',
    'capAdd',
    'securityOpt',
    'remoteUser',
    'userEnvProbe',
    'remoteEnv',
    'overrideCommand',
    'portsAttributes',
    'otherPortsAttributes',
    'forwardPorts',
    'shutdownAction',
    'updateRemoteUserUID',
    'hostRequirements',
];
const pickFeatureProperties = [
    'id',
    'init',
    'privileged',
    'capAdd',
    'securityOpt',
    'entrypoint',
    'mounts',
    'customizations',
];
const replaceProperties = [
    'customizations',
    'entrypoint',
    'onCreateCommand',
    'updateContentCommand',
    'postCreateCommand',
    'postStartCommand',
    'postAttachCommand',
    'shutdownAction'
];
function mergeConfiguration(config, imageMetadata) {
    var _a, _b, _c, _d, _e, _f, _g, _h;
    const customizations = imageMetadata.reduce((obj, entry) => {
        for (const key in entry.customizations) {
            if (key in obj) {
                obj[key].push(entry.customizations[key]);
            }
            else {
                obj[key] = [entry.customizations[key]];
            }
        }
        return obj;
    }, {});
    const reversed = imageMetadata.reverse();
    const copy = { ...config };
    replaceProperties.forEach(property => delete copy[property]);
    const merged = {
        ...copy,
        init: imageMetadata.some(entry => entry.init),
        privileged: imageMetadata.some(entry => entry.privileged),
        capAdd: unionOrUndefined(imageMetadata.map(entry => entry.capAdd)),
        securityOpt: unionOrUndefined(imageMetadata.map(entry => entry.securityOpt)),
        entrypoints: collectOrUndefined(imageMetadata, 'entrypoint'),
        mounts: concatOrUndefined(imageMetadata.map(entry => entry.mounts)),
        customizations: Object.keys(customizations).length ? customizations : undefined,
        onCreateCommands: collectOrUndefined(imageMetadata, 'onCreateCommand'),
        updateContentCommands: collectOrUndefined(imageMetadata, 'updateContentCommand'),
        postCreateCommands: collectOrUndefined(imageMetadata, 'postCreateCommand'),
        postStartCommands: collectOrUndefined(imageMetadata, 'postStartCommand'),
        postAttachCommands: collectOrUndefined(imageMetadata, 'postAttachCommand'),
        waitFor: (_a = reversed.find(entry => entry.waitFor)) === null || _a === void 0 ? void 0 : _a.waitFor,
        remoteUser: (_b = reversed.find(entry => entry.remoteUser)) === null || _b === void 0 ? void 0 : _b.remoteUser,
        containerUser: (_c = reversed.find(entry => entry.containerUser)) === null || _c === void 0 ? void 0 : _c.containerUser,
        userEnvProbe: (_d = reversed.find(entry => entry.userEnvProbe)) === null || _d === void 0 ? void 0 : _d.userEnvProbe,
        remoteEnv: Object.assign({}, ...imageMetadata.map(entry => entry.remoteEnv)),
        containerEnv: Object.assign({}, ...imageMetadata.map(entry => entry.containerEnv)),
        overrideCommand: (_e = reversed.find(entry => typeof entry.overrideCommand === 'boolean')) === null || _e === void 0 ? void 0 : _e.overrideCommand,
        portsAttributes: Object.assign({}, ...imageMetadata.map(entry => entry.portsAttributes)),
        otherPortsAttributes: (_f = reversed.find(entry => entry.otherPortsAttributes)) === null || _f === void 0 ? void 0 : _f.otherPortsAttributes,
        forwardPorts: mergeForwardPorts(imageMetadata),
        shutdownAction: (_g = reversed.find(entry => entry.shutdownAction)) === null || _g === void 0 ? void 0 : _g.shutdownAction,
        updateRemoteUserUID: (_h = reversed.find(entry => entry.updateRemoteUserUID)) === null || _h === void 0 ? void 0 : _h.updateRemoteUserUID,
        hostRequirements: mergeHostRequirements(imageMetadata),
    };
    return merged;
}
exports.mergeConfiguration = mergeConfiguration;
function mergeForwardPorts(imageMetadata) {
    const forwardPorts = [
        ...new Set([].concat(...imageMetadata.map(entry => entry.forwardPorts || []))
            .map(port => typeof port === 'number' ? `localhost:${port}` : port))
    ].map(port => /localhost:\d+/.test(port) ? parseInt(port.substring('localhost:'.length)) : port);
    return forwardPorts.length ? forwardPorts : undefined;
}
function mergeHostRequirements(imageMetadata) {
    const cpus = Math.max(...imageMetadata.map(m => { var _a; return ((_a = m.hostRequirements) === null || _a === void 0 ? void 0 : _a.cpus) || 0; }));
    const memory = Math.max(...imageMetadata.map(m => { var _a; return parseBytes(((_a = m.hostRequirements) === null || _a === void 0 ? void 0 : _a.memory) || '0'); }));
    const storage = Math.max(...imageMetadata.map(m => { var _a; return parseBytes(((_a = m.hostRequirements) === null || _a === void 0 ? void 0 : _a.storage) || '0'); }));
    return cpus || memory || storage ? {
        cpus,
        memory: memory ? `${memory}` : undefined,
        storage: storage ? `${storage}` : undefined,
    } : undefined;
}
function parseBytes(str) {
    const m = /^(\d+)([tgmk]b)?$/.exec(str);
    if (m) {
        const [, strn, stru] = m;
        const n = parseInt(strn, 10);
        const u = stru && { t: 2 ** 40, g: 2 ** 30, m: 2 ** 20, k: 2 ** 10 }[stru[0]] || 1;
        return n * u;
    }
    return 0;
}
function concatOrUndefined(entries) {
    const values = [].concat(...entries.filter(entry => !!entry));
    return values.length ? values : undefined;
}
function unionOrUndefined(entries) {
    const values = [...new Set([].concat(...entries.filter(entry => !!entry)))];
    return values.length ? values : undefined;
}
function collectOrUndefined(entries, property) {
    const values = entries.map(entry => entry[property])
        .filter(value => !!value);
    return values.length ? values : undefined;
}
function getDevcontainerMetadata(baseImageMetadata, devContainerConfig, featuresConfig) {
    const features = Array.isArray(featuresConfig) ? featuresConfig : [].concat(...featuresConfig.featureSets
        .map(x => x.features));
    const raw = features.map(feature => pick(feature, pickFeatureProperties));
    return {
        config: [
            ...baseImageMetadata.config,
            ...raw.map(devContainerConfig.substitute),
            pick(devContainerConfig.config, pickConfigProperties),
        ].filter(config => Object.keys(config).length),
        raw: [
            ...baseImageMetadata.raw,
            ...raw,
            pick(devContainerConfig.raw, pickConfigProperties),
        ].filter(config => Object.keys(config).length),
        substitute: devContainerConfig.substitute,
    };
}
exports.getDevcontainerMetadata = getDevcontainerMetadata;
function pick(obj, keys) {
    return keys.reduce((res, key) => {
        if (key in obj) {
            res[key] = obj[key];
        }
        return res;
    }, {});
}
async function getImageBuildInfo(params, configWithRaw, experimentalImageMetadata) {
    var _a, _b;
    const { dockerCLI, dockerComposeCLI } = params;
    const { cliHost, output } = 'cliHost' in params ? params : params.common;
    const { config } = configWithRaw;
    if ((0, configuration_1.isDockerFileConfig)(config)) {
        const dockerfileUri = (0, configuration_1.getDockerfilePath)(cliHost, config);
        const dockerfilePath = await (0, utils_1.uriToWSLFsPath)(dockerfileUri, cliHost);
        if (!cliHost.isFile(dockerfilePath)) {
            throw new errors_1.ContainerError({ description: `Dockerfile (${dockerfilePath}) not found.` });
        }
        const dockerfile = (await cliHost.readFile(dockerfilePath)).toString();
        return getImageBuildInfoFromDockerfile(params, dockerfile, ((_a = config.build) === null || _a === void 0 ? void 0 : _a.args) || {}, (_b = config.build) === null || _b === void 0 ? void 0 : _b.target, configWithRaw.substitute, experimentalImageMetadata);
    }
    else if ('dockerComposeFile' in config) {
        const cwdEnvFile = cliHost.path.join(cliHost.cwd, '.env');
        const envFile = Array.isArray(config.dockerComposeFile) && config.dockerComposeFile.length === 0 && await cliHost.isFile(cwdEnvFile) ? cwdEnvFile : undefined;
        const composeFiles = await (0, configuration_1.getDockerComposeFilePaths)(cliHost, config, cliHost.env, cliHost.cwd);
        const buildParams = { cliHost, dockerCLI, dockerComposeCLI, env: cliHost.env, output };
        const composeConfig = await (0, dockerCompose_1.readDockerComposeConfig)(buildParams, composeFiles, envFile);
        const services = Object.keys(composeConfig.services || {});
        if (services.indexOf(config.service) === -1) {
            throw new Error(`Service '${config.service}' configured in devcontainer.json not found in Docker Compose configuration.`);
        }
        const composeService = composeConfig.services[config.service];
        const serviceInfo = (0, dockerCompose_1.getBuildInfoForService)(composeService, cliHost.path, composeFiles);
        if (serviceInfo.build) {
            const { context, dockerfilePath } = serviceInfo.build;
            const resolvedDockerfilePath = cliHost.path.isAbsolute(dockerfilePath) ? dockerfilePath : cliHost.path.resolve(context, dockerfilePath);
            const dockerfile = (await cliHost.readFile(resolvedDockerfilePath)).toString();
            return getImageBuildInfoFromDockerfile(params, dockerfile, serviceInfo.build.args || {}, serviceInfo.build.target, configWithRaw.substitute, experimentalImageMetadata);
        }
        else {
            return getImageBuildInfoFromImage(params, composeService.image, configWithRaw.substitute, experimentalImageMetadata);
        }
    }
    else {
        return getImageBuildInfoFromImage(params, config.image, configWithRaw.substitute, experimentalImageMetadata);
    }
}
exports.getImageBuildInfo = getImageBuildInfo;
async function getImageBuildInfoFromImage(params, imageName, substitute, experimentalImageMetadata) {
    const imageDetails = await (0, utils_1.inspectDockerImage)(params, imageName, true);
    const user = imageDetails.Config.User || 'root';
    const { output } = 'output' in params ? params : params.common;
    const metadata = getImageMetadata(imageDetails, substitute, experimentalImageMetadata, output);
    return {
        user,
        metadata,
        imageDetails,
    };
}
exports.getImageBuildInfoFromImage = getImageBuildInfoFromImage;
async function getImageBuildInfoFromDockerfile(params, dockerfile, dockerBuildArgs, targetStage, substitute, experimentalImageMetadata) {
    const { output } = 'output' in params ? params : params.common;
    return internalGetImageBuildInfoFromDockerfile(imageName => (0, utils_1.inspectDockerImage)(params, imageName, true), dockerfile, dockerBuildArgs, targetStage, substitute, experimentalImageMetadata, output);
}
exports.getImageBuildInfoFromDockerfile = getImageBuildInfoFromDockerfile;
async function internalGetImageBuildInfoFromDockerfile(inspectDockerImage, dockerfile, dockerBuildArgs, targetStage, substitute, experimentalImageMetadata, output) {
    const extracted = (0, dockerfileUtils_1.extractDockerfile)(dockerfile);
    const dockerfileUser = (0, dockerfileUtils_1.findUserStatement)(extracted, dockerBuildArgs, targetStage);
    const baseImage = (0, dockerfileUtils_1.findBaseImage)(extracted, dockerBuildArgs, targetStage);
    const imageDetails = baseImage && await inspectDockerImage(baseImage) || undefined;
    const user = dockerfileUser || (imageDetails === null || imageDetails === void 0 ? void 0 : imageDetails.Config.User) || 'root';
    const metadata = imageDetails ? getImageMetadata(imageDetails, substitute, experimentalImageMetadata, output) : { config: [], raw: [], substitute };
    return {
        user,
        metadata,
    };
}
exports.internalGetImageBuildInfoFromDockerfile = internalGetImageBuildInfoFromDockerfile;
exports.imageMetadataLabel = 'devcontainer.metadata';
function getImageMetadataFromContainer(containerDetails, devContainerConfig, featuresConfig, experimentalImageMetadata, output) {
    if (!experimentalImageMetadata) {
        return getDevcontainerMetadata({ config: [], raw: [], substitute: devContainerConfig.substitute }, devContainerConfig, featuresConfig);
    }
    return internalGetImageMetadata(containerDetails, devContainerConfig.substitute, experimentalImageMetadata, output);
}
exports.getImageMetadataFromContainer = getImageMetadataFromContainer;
function getImageMetadata(imageDetails, substitute, experimentalImageMetadata, output) {
    return internalGetImageMetadata(imageDetails, substitute, experimentalImageMetadata, output);
}
exports.getImageMetadata = getImageMetadata;
function internalGetImageMetadata(imageDetails, substitute, experimentalImageMetadata, output) {
    const raw = internalGetImageMetadata0(imageDetails, experimentalImageMetadata, output);
    return {
        config: raw.map(substitute),
        raw,
        substitute,
    };
}
function internalGetImageMetadata0(imageDetails, experimentalImageMetadata, output) {
    if (!experimentalImageMetadata) {
        return [];
    }
    const str = (imageDetails.Config.Labels || {})[exports.imageMetadataLabel];
    if (str) {
        try {
            const obj = JSON.parse(str);
            if (Array.isArray(obj)) {
                return obj;
            }
            if (obj && typeof obj === 'object') {
                return [obj];
            }
            output.write(`Invalid image metadata: ${str}`);
        }
        catch (err) {
            output.write(`Error parsing image metadata: ${(err === null || err === void 0 ? void 0 : err.message) || err}`);
        }
    }
    return [];
}
function getDevcontainerMetadataLabel(baseImageMetadata, devContainerConfig, featuresConfig, experimentalImageMetadata) {
    if (!experimentalImageMetadata) {
        return '';
    }
    const metadata = getDevcontainerMetadata(baseImageMetadata, devContainerConfig, featuresConfig).raw;
    if (!metadata.length) {
        return '';
    }
    const imageMetadataLabelValue = metadata.length !== 1
        ? `[${metadata
            .map(feature => ` \\\n${toLabelString(feature)}`)
            .join(',')} \\\n]`
        : toLabelString(metadata[0]);
    return `LABEL ${exports.imageMetadataLabel}="${imageMetadataLabelValue}"`;
}
exports.getDevcontainerMetadataLabel = getDevcontainerMetadataLabel;
function toLabelString(obj) {
    return JSON.stringify(obj)
        .replace(/(?=["\\$])/g, '\\');
}
//# sourceMappingURL=imageMetadata.js.map